# Rubik's Cubes 3x3, 5x5 and 8x8
# using serial ws2812 color leds
# in MicroPython on Raspberry Pi Pico

# RubiksCubesLeds.py

# version 1.2 2023-08-01

# MIT license https://opensource.org/license/mit/

# Copyright (c) 2023 Ronald Lokker      ronald@lokker.net
# Copyright (c) 2023 Theo van Haandel   rubiksled555@x4a.nl
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
# associated documentation files (the “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial
# portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# parts based on examples in Raspberry Pi Pico SDK
# https://www.raspberrypi.com/documentation/microcontrollers/raspberry-pi-pico.html

import array
import time
try:
    import machine
    import rp2
except ModuleNotFoundError:
    print('Not running on Raspberry Pi Pico')
    rp2 = None

black  =   0,   0,   0
red    = 255,   0,   0
yellow = 255, 255,   0
green  =   0, 255,   0
cyan   =   0, 255, 255
blue   =   0,   0, 255
purple = 255,   0, 255
white  = 255, 255, 255
orange = 255, 100,   0

colors = black, red, yellow, green, cyan, blue, purple, white, orange

dimmer =  32 # 1000
def dim(color):
    return color // dimmer

if rp2:
    @rp2.asm_pio(sideset_init=rp2.PIO.OUT_LOW, out_shiftdir=rp2.PIO.SHIFT_LEFT, autopull=True, pull_thresh=24)
    def ws2812():
        wrap_target()
        label("loop")
        out(x, 1)               .side(0)    [2]
        jmp(not_x, "low")       .side(1)    [1]
        jmp("loop")             .side(1)    [4]
        label("low")
        nop()                   .side(0)    [4]
        wrap()

picoPin = 16

class LedMachine:  # uses pio, only single instance
    def __init__(self):
        if rp2:
            self.sm = rp2.StateMachine(0, ws2812, freq=8_000_000, sideset_base=machine.Pin(picoPin))
            self.sm.active(1)
    def set(self, rgbColors):  # rgb tuples
        def grb(r, g, b):
             return (dim(g) << 16) + (dim(r) << 8) + dim(b)
        ar = array.array("I", [grb(*rgbColors[idx]) for idx in range(len(rgbColors))])
        if rp2:
            self.sm.put(ar, 8)

cubeSideCount = 6

class LedCube:  # physical color led cube
    # original Rubik's cube colors
    #      white
    # green red blue orange 
    #      yellow
    # match physical cube to original
    # in order    top, left, front, right, back, down
    cubeColors = white, green, red, blue, orange, yellow

    def __init__(self, ledRowCount, ledMachine = LedMachine()):
        self.ledRowCount = ledRowCount
        if ledRowCount == 5:  # original led cube
            self.sidesMap = 5, 2, 3, 4, 1, 0
            # rotate side 0 yellow 270, side 5 white 90
            self.sidesRotate = 270, 0, 0, 0, 0, 90
            self.zigZag = True
        elif ledRowCount == 3:
            self.sidesMap = 0, 1, 2, 3, 4, 5 # todo: match real led cube
            self.sidesRotate = 0, 0, 0, 0, 0, 0 # todo
            self.zigZag = False
        elif ledRowCount == 8:
            self.sidesMap = 0, 1, 2, 3, 4, 5 # todo: match real led cube
            self.sidesRotate = 0, 0, 0, 0, 0, 0 # todo
            self.zigZag = True
        else:
            # todo: accept different types, sidesMap rotate zigzag in type text
            raise ValueError('ledRowCount ' + str(ledRowCount) +
                             ' not implemented')
        self.ledSideCount = self.ledRowCount * self.ledRowCount
        self.ledCubeCount = cubeSideCount * self.ledSideCount
        self.ledMachine = ledMachine

    def ledZigZagOffset(self, sideIdx, idx):
        if self.zigZag:
            row, col = divmod(idx, self.ledRowCount)
            if row % 2:
                idx = (row + 1) * self.ledRowCount - col - 1
        return self.sidesMap[sideIdx] * self.ledSideCount + idx

    # see also rotate in RubiksCube
    # https://stackoverflow.com/questions/42519/how-do-you-rotate-a-two-dimensional-array
    def ledRotateZigZagOffset(self, sideIdx, idx):
        rotate = self.sidesRotate[sideIdx]
        if rotate == 90:
            row, col = divmod(idx, self.ledRowCount)
            idx = col * self.ledRowCount + self.ledRowCount - row - 1
        elif rotate == 180:
            row, col = divmod(idx, self.ledRowCount)
            idx = (self.ledRowCount - col - 1) * self.ledRowCount - row - 1
        elif rotate == 270:
            row, col = divmod(idx, self.ledRowCount)
            idx = (self.ledRowCount - col - 1) * self.ledRowCount + row
        # else 0 do not rotate
        return self.ledZigZagOffset(sideIdx, idx)

    def clearLeds(self):
        print('clearLeds')
        ledColors = [ black ] * self.ledCubeCount
        self.ledMachine.set(ledColors)

    def testAllLeds(self):
        # set all sides to single color one after another to test leds
        for color in colors:
            print('testAllLeds', color)
            ledColors = [ color ] * self.ledCubeCount
            self.ledMachine.set(ledColors)
            # time.sleep(0.5)

    def testSingleLeds(self):
        # set single led green one after another to detect order like zigzag
        ledColors = [ black ] * self.ledCubeCount
        for idx in range(self.ledCubeCount):
            print('testSingleLeds', idx, self.ledCubeCount)
            ledColors[idx] = green
            self.ledMachine.set(ledColors)
            ledColors[idx] = black
            # time.sleep(0.1)

    def testLinearLeds(self):
        # zigzag compensated
        print('testLinearLeds')
        ledColors = [ black ] * self.ledCubeCount
        for sideIdx in range(cubeSideCount):
            for idx in range(self.ledSideCount):
                ledColor = self.cubeColors[sideIdx]
                if idx == 0 or idx == sideIdx:
                    ledColor = black
                idx = self.ledZigZagOffset(sideIdx, idx)
                ledColors[idx] = ledColor
        self.ledMachine.set(ledColors)

    def testCubeLeds(self):
        # first led compensated
        print('testCubeLeds')
        ledColors = [ black ] * self.ledCubeCount
        for sideIdx in range(cubeSideCount):
            for idx in range(self.ledSideCount):
                ledColor = self.cubeColors[sideIdx]
                if idx == 0:
                    ledColor = black
                idx = self.ledRotateZigZagOffset(sideIdx, idx)
                ledColors[idx] = ledColor
        self.ledMachine.set(ledColors)

    # todo: cache rubikscube to led cube side row col mappings
    def setRubiksLeds(self, rubiksCube = None):
        # map led row to rubikscube row
        # led cube is 3x3, 5x5, 8x8, rubiks cube can be 2x2 to 8x8
        rubiksIndex = rubiksCube.rowCount - 2
        if self.ledRowCount == 3:
            rowMap = ( (0, None, 1), # 2x2
                       (0,    1, 2), # 3x3
                     )[rubiksIndex]
        elif self.ledRowCount == 5:
            rowMap = ( (0,    0, None,    1, 1), # 2x2
                       (0, None,    1, None, 2), # 3x3
                     # (0,    0,    1,    2, 2), # 3x3
                       (0,    1, None,    2, 3), # 4x4
                       (0,    1,    2,    3, 4), # 5x5
                     )[rubiksIndex]
        elif self.ledRowCount == 8:
            rowMap = ( (0,    0,    0,    0,   1,    1,    1, 1), # 2x2
                       (0,    0, None,    1,   1, None,    2, 2), # 3x3
                       (0,    0,    1,    1,   2,    2,    3, 3), # 4x4
                       (0, None,    1,    2,   2,    3, None, 4), # 5x5
                       (0,    0,    1,    2,   3,    4,    5, 5), # 6x6
                       (0,    1,    2,    3,   3,    4,    5, 6), # 7x7
                       (0,    1,    2,    3,   4,    5,    6, 7), # 8x8
                     )[rubiksIndex]
        else:
            raise ValueError('ledRowCount is ' + str(ledRowCount) +
                             ', only 3, 5 and 8 implemented')
        ledColors = [ black for ix in range(self.ledCubeCount) ]
        for sideIdx in range(cubeSideCount):
            for ledRow in range(self.ledRowCount):
                for ledCol in range(self.ledRowCount):
                    ledColor = black
                    cubeRow, cubeCol = rowMap[ledRow], rowMap[ledCol]
                    if cubeRow is not None and cubeCol is not None:
                        ledColor = self.cubeColors[rubiksCube.sides[sideIdx][cubeRow][cubeCol]]
                    idx = ledRow * self.ledRowCount + ledCol
                    idx = self.ledRotateZigZagOffset(sideIdx, idx)
                    ledColors[idx] = ledColor
        self.ledMachine.set(ledColors)

up, left, front, right, back, down = range(cubeSideCount)

north, east, south, west = range(4)

class RubiksCube:  # virtual cube memory model
    def __init__(self, rowCountOrStr):
        if isinstance(rowCountOrStr, int):
            self.rowCount = rowCountOrStr
            self.reset()
        else: # reverse of __str__
            lines = [ line.strip() for line in rowCountOrStr.split('\n')
                      if line != '' ]
            self.rowCount = len(lines) // 3
            self.sides = [ [ [ ] for row in range(self.rowCount) ]
                           for _ in range(cubeSideCount) ]
            for rowIdx in range(self.rowCount):
                line = lines[rowIdx] + lines[rowIdx + 3] + lines[rowIdx + 6]
                sideIdx = 0
                colIdx = 0
                for ch in line:
                    if ch.isdigit():
                        self.sides[sideIdx][rowIdx].append(int(ch))
                        colIdx += 1
                        if colIdx == self.rowCount:
                            colIdx = 0
                            sideIdx += 1
        # print(self.sides)

    def __str__(self):
        strSides = [ [ ''.join(str(col) for col in row) for row in side ]
                     for side in self.sides ]
        # print(strSides)
        res = ''
        for layer in ( ( None, up ),
                       ( left, front, right, back ),
                       ( None, down ) ):
            for rowIndex in range(self.rowCount):
                for sideIndex in layer:
                    if sideIndex is None:
                        res += ' ' * self.rowCount
                    else:
                        res += strSides[sideIndex][rowIndex]
                    res += ' '
                res += '\n'
        return res
    def __eq__(self, other):
        return self.sides == other.sides
    def reset(self):
        self.sides = [ [ [ x for col in range(self.rowCount) ]
                         for row in range(self.rowCount) ]
                       for x in range(cubeSideCount) ]
    # see also rotate in LedCube
    # see also https://stackoverflow.com/questions/8372399/zip-with-list-output-instead-of-tuple
    def sideRotate90(self, sideIdx):
        self.sides[sideIdx] = [ list(t) for t in zip(*reversed(self.sides[sideIdx])) ]
    def sideRotate180(self, sideIdx):
        self.sideRotate90(sideIdx)
        self.sideRotate90(sideIdx)
    def sideRotate270(self, sideIdx):
        self.sides[sideIdx] = list(reversed([ list(t) for t in zip(*self.sides[sideIdx]) ]))
    def rowRotate90(self, sideIndex, rowIter = (0,)):
        parms = ( ( (left, north), (front, north), (right, north), (back,  north) ),
                  ( (up,   west),  (back,  east),  (down,  west),  (front, west) ),
                  ( (up,   south), (left,  east),  (down,  north), (right, west) ),
                  ( (up,   east),  (front, east),  (down,  east),  (back,  west) ),
                  ( (up,   north), (right, east),  (down,  south), (left,  west) ),
                  ( (left, south), (back,  south), (right, south), (front, south) ) )[sideIndex]
        for rowIndex in rowIter:
            def sideGetRow(pos):
                sideIdx, Nesw = pos
                rowIdx = rowIndex if Nesw in (north, west) else -1 - rowIndex
                if Nesw in (north, south):
                    row = self.sides[sideIdx][rowIdx]
                else:
                    row = [row[rowIdx] for row in self.sides[sideIdx]]
                if Nesw in (south, west):
                    row = list(reversed(row))
                return row
            def sideSetRow(pos, row):
                sideIdx, Nesw = pos
                rowIdx = rowIndex if Nesw in (north, west) else -1 - rowIndex
                if Nesw in (south, west):
                    row = list(reversed(row))
                if Nesw in (north, south):
                    self.sides[sideIdx][rowIdx] = row
                else:
                    for row, val in zip(self.sides[sideIdx], row):
                        row[rowIdx] = val
            tempRow = sideGetRow(parms[0])
            sideSetRow(parms[0], sideGetRow(parms[1]))
            sideSetRow(parms[1], sideGetRow(parms[2]))
            sideSetRow(parms[2], sideGetRow(parms[3]))
            sideSetRow(parms[3], tempRow)
    def rowRotate180(self, sideIdx):
        self.rowRotate90(sideIdx)
        self.rowRotate90(sideIdx)
    def rowRotate270(self, sideIdx):
        self.rowRotate90(sideIdx)
        self.rowRotate180(sideIdx)
    def rotate90(self, sideIdx):
        self.rowRotate90(sideIdx)
        self.sideRotate90(sideIdx)
    def rotate180(self, sideIdx):
        self.rowRotate180(sideIdx)
        self.sideRotate180(sideIdx)
    def rotate270(self, sideIdx):
        self.rowRotate270(sideIdx)
        self.sideRotate270(sideIdx)
    def turn(self, turnTxt):
        turns = turnTxt.lower().split()
        for turn in turns:
            turnCh = turn[0]
            turnIdx = 'ulfrbd'.index(turnCh)
            if len(turn) == 1:
                self.rotate90(turnIdx)
            elif turn[1] == '2':
                self.rotate180(turnIdx)
            elif turn[1] == "'":
                self.rotate270(turnIdx)

def testRubik():
    cube1 = RubiksCube(3)
    # cube1.rotate90(2)
    print(cube1)
    # return

    txtCube1 = str(cube1)
    print(txtCube1)
    cube2 = RubiksCube(txtCube1)
    print(cube2)
    print(cube1 == cube2)


def testRubiksCubeLeds(ledRowCount):
    ledCube = LedCube(ledRowCount)
    ledCube.clearLeds()
    # time.sleep(5)
    for cubeRowCount in range(2, ledRowCount + 1):
        rubiksCube = RubiksCube(cubeRowCount)
        for idx in range(10):
            ledCube.setRubiksLeds(rubiksCube)
            print(rubiksCube)
            # time.sleep(0.2)
            # input("1>")
            # rubiksCube.rowRotate90(0)
            rubiksCube.rotate90(0)
            ledCube.setRubiksLeds(rubiksCube)
            print(rubiksCube)
            # time.sleep(0.2)
            # input("2>")
            # rubiksCube.sideRotate90(2)
            rubiksCube.rotate270(2)
            ledCube.setRubiksLeds(rubiksCube)
            print(rubiksCube)
            # time.sleep(0.2)
            # input("3>")
            # rubiksCube.sideRotate90(2)
            rubiksCube.rotate180(3)

def testSimple(ledRowCount):
    ledCube = LedCube(ledRowCount)
    ledCube.clearLeds()
    for rubiksRowCount in range(2, ledRowCount + 1):
        rubiksCube = RubiksCube(rubiksRowCount)
        ledCube.setRubiksLeds(rubiksCube)
        print(rubiksCube)
        sideIdx = front # 2
        for idx in range(4):
            # time.sleep(0.2)
            rubiksCube.rotate90(sideIdx)
            rubiksCube.rowRotate90(sideIdx, range(1))
            rubiksCube.rowRotate90(sideIdx, range(2))
            if rubiksRowCount >= 3:
                rubiksCube.rowRotate90(sideIdx, range(3))
            if rubiksRowCount >= 4:
                rubiksCube.rowRotate90(sideIdx, range(4))
            rubiksCube.rowRotate90(sideIdx, range(1, 2))
            if rubiksRowCount >= 3:
                rubiksCube.rowRotate90(sideIdx, range(1, 3))
            if rubiksRowCount >= 4:
                rubiksCube.rowRotate90(sideIdx, range(1, 4))
            if rubiksRowCount >= 3:
                rubiksCube.rowRotate90(sideIdx, range(2, 3))
            if rubiksRowCount >= 4:
                rubiksCube.rowRotate90(sideIdx, range(2, 4))
                rubiksCube.rowRotate90(sideIdx, range(3, 4))
            if rubiksRowCount >= 4:
                rubiksCube.rowRotate90(sideIdx, (1, 3))
            ledCube.setRubiksLeds(rubiksCube)
            print(rubiksCube)

def testTurns(ledRowCount):
    ledCube = LedCube(ledRowCount)
    ledCube.clearLeds()
    for rubiksRowCount in range(2, ledRowCount + 1):
        rubiksCube = RubiksCube(rubiksRowCount)
        # ledCube.setRubiksLeds(rubiksCube)
        # print(rubiksCube)
        if False:
            rubiksCube.turn('u')
            ledCube.setRubiksLeds(rubiksCube)
            print(rubiksCube)
            rubiksCube.turn('f2')
            ledCube.setRubiksLeds(rubiksCube)
            print(rubiksCube)
        # rubiksCube.turn('u f2')
        # ledCube.setRubiksLeds(rubiksCube)
        # print(rubiksCube)
        rubiksCube.turn('u f2')
        ledCube.setRubiksLeds(rubiksCube)
        print(rubiksCube)
        rubiksCube.reset()
        rubiksCube.turn("u' l' u l u f u' f'")
        ledCube.setRubiksLeds(rubiksCube)
        print(rubiksCube)
    
def main():
    if True: # while True:
        ledRowCounts = 5, 3, 8
        if False: # True:
            for ledRowCount in ledRowCounts:
                ledCube = LedCube(ledRowCount)
                ledCube.clearLeds()
                ledCube.testAllLeds()
                ledCube.testSingleLeds()
                ledCube.testLinearLeds()
                ledCube.testCubeLeds()
        testRubik()
        if False: # True:
            for ledRowCount in ledRowCounts:
                testRubiksCubeLeds(ledRowCount)
                testSimple(ledRowCount)
        for ledRowCount in ledRowCounts:
            testTurns(ledRowCount)
    # in main.py
    # while True:
    #    testRubiksCubeLeds()

if __name__ == '__main__':
    main()
